/*  Kernel module help for ia64.
    Copyright (C) 2002 Rusty Russell.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <linux/moduleloader.h>
#include <linux/elf.h>
#include <linux/vmalloc.h>
#include <linux/err.h>
#include <linux/string.h>
#include <linux/kernel.h>

#ifndef __KERNEL__
#include <sys/mman.h>
#include <unistd.h>
#endif

#if 1
#define DEBUGP printk
#else
#define DEBUGP(fmt...)
#endif

/* IA64 function descriptors look like this. */
struct ia64_fdesc
{
	u64 entry;	/* code entry point */
	u64 gp;		/* global pointer */
};

struct ia64_callout_stub
{
	/* Three instruction bundles in PLT. */
 	unsigned char plt[16 * 3];
};

/* Confusingly, the pltoff section referred to by the (generated)
   .plt, expects descriptors which look like this: */
struct ia64_plt
{
	u64 gp;		/* global pointer */
	u64 entry;	/* code entry point */
};

/* We need to create a function descriptor for any internal function
   which is referenced.  ld.so uses the size of the dynamic symbol
   table to allocated them: it does this by assuming the DT_STRTAB
   immediately follows the DT_SYMTAB.

   THIS IS WHY THE FUCKING LINKER SHOULD CREATE THESE FUCKING
   FUNCTION DESCRIPTORS FOR US LIKE A SANE FUCKING ARCHITECTURE --RR */
static unsigned long num_fdescs(Elf_Ehdr *hdr, Elf_Dyn *dyn)
{
	u64 i, symtaboff = 0, strtaboff = 0;

	/* Missing sections get caught in the main code later. */
	for (i = 0; dyn[i].d_tag != DT_NULL; i++) {
		if (dyn[i].d_tag == DT_SYMTAB)
			symtaboff = dyn[i].d_un.d_ptr;
		else if (dyn[i].d_tag == DT_STRTAB)
			strtaboff = dyn[i].d_un.d_ptr;
	}
	return (strtaboff - symtaboff) / sizeof(Elf64_Sym);
}

void *module_alloc(Elf_Ehdr *hdr, Elf_Dyn *dyn, unsigned long modsize,
		   unsigned long *extra)
{
	void *ptr;

	*extra = num_fdescs(hdr, dyn) * sizeof(struct ia64_fdesc);
	ptr = vmalloc(*extra + modsize);
	if (!ptr)
		return ERR_PTR(-ENOMEM);

#ifndef __KERNEL__
	{
		long off = ((long) ptr & (getpagesize() - 1));
		mprotect((void *) ((long)ptr - off),
			 modsize + *extra + off,
			 PROT_READ|PROT_WRITE|PROT_EXEC);
	}
#endif
	return ptr;
}

void module_free(void *modmem)
{
	vfree(modmem);
}

static struct ia64_fdesc *find_fdesc(struct module *me, u64 value)
{
	struct ia64_fdesc *fdesc = me->loadaddr - me->extra;

	/* Find or generate the (unique) function descriptor. */
	while ((void *)fdesc < me->loadaddr) {
		if (fdesc->entry == value)
			return fdesc;
		if (fdesc->entry == 0) {
			fdesc->entry = value;
			fdesc->gp = me->arch.got;
			return fdesc;
		}
		fdesc++;
	}
	BUG();
	return NULL;
}

static u64 calculate_got(struct module *owner)
{
	u64 got;
	if (!owner) {
		/* Kernel's GOT value = our current gp. */
		asm("mov %0 = gp" : "=r" (got));
		DEBUGP("GOT = %lu\n", got);
	} else
		got = owner->arch.got;
	return got;
}

static void fixup_plt(struct ia64_plt *plt, u64 value)
{
	struct ia64_fdesc *fdesc = (void *)value;
	plt->gp = fdesc->gp;
	plt->entry = fdesc->entry;
}

int apply_relocate(Elf_Rel *rel,
		   unsigned int num_rel,
		   Elf_Dyn *dyn,
		   Elf_Sym *syms,
		   const char *strtab,
		   struct module *me)
{
	printk(KERN_ERR "module %s: non-ADD RELOCATION unsupported\n", me->name);
	return -ENOEXEC;
}

/* Apply the given add relocations.  Return -error or 0 */
int apply_relocate_add(Elf_Rela *rela,
		       unsigned int num_rela,
		       Elf_Dyn *dyn,
		       Elf_Sym *syms,
		       const char *strtab,
		       struct module *me)
{
	unsigned int i;
	struct module *owner;

	/* Find out our GOT. */
	for (i = 0; dyn[i].d_tag != DT_PLTGOT; i++) {
		if (dyn[i].d_tag == DT_NULL) {
			printk("%s: No DT_PLTGOT\n", me->name);
			return -ENOEXEC;
		}
	}
	me->arch.got = (u64)me->loadaddr + dyn[i].d_un.d_ptr;

	for (i = 0; i < num_rela; i++) {
		unsigned long r_type = ELF64_R_TYPE(rela[i].r_info);
		u64 *location = (void *)me->loadaddr + rela[i].r_offset;
		unsigned int symnum;
		unsigned long value = 0;

		/* This is the symbol it is referring to.  */
		symnum = ELF64_R_SYM(rela[i].r_info);
		if (symnum) {
			value = (u64)resolve_symbol(me, symnum, syms, strtab,
						    &owner);
			if (!value)
				return -ENOENT;
		}
		value += rela[i].r_addend;

		switch (r_type) {
		case R_IA64_REL64LSB:
			DEBUGP("location %p (got + %li) = %p\n", location,
			       (unsigned long)location - me->arch.got,
			       me->loadaddr + rela[i].r_addend);
			*location = (u64)me->loadaddr + rela[i].r_addend;
			break;

		case R_IA64_FPTR64LSB:
			if (owner != me)
				*location = value;
			else
				*location = (u64)find_fdesc(me, value);
			DEBUGP("fdesc %p for %p\n", (void *)value,
			       (void *)*location);
			break;

		case R_IA64_IPLTLSB:
			printk("PLT location %p -> %p\n", location,
			       (void *)value);
			fixup_plt((void *)location, value);
			break;

		default:
			printk(KERN_ERR "module %s: Unknown relocation: %lu\n",
			       me->name, r_type);
			return -ENOEXEC;
		}
	}

	return 0;
}

int module_finalize(Elf_Ehdr *hdr, struct module *me)
{
	return 0;
}

void module_truncate(struct module *me)
{
}
