/* Kernel module help for PPC64
   Copyright (C) 2002  Rusty Russell, IBM.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  */

#include <linux/moduleloader.h>
#include <linux/elf.h>
#include <linux/vmalloc.h>
#include <linux/err.h>
#include <linux/string.h>
#include <linux/kernel.h>

#if 1
#define DEBUGP printk
#else
#define DEBUGP(fmt , ...)
#endif

void *module_alloc(Elf_Ehdr *hdr, Elf_Dyn *dyn, unsigned long size,
		   unsigned long *extra)
{
	return (vmalloc(size) ?: ERR_PTR(-ENOMEM));
}

void
module_free(void *region)
{
	vfree(region);
}

/* Apply the given relocations.  Return -error or 0. */
int apply_relocate(Elf_Rel *rel,
		   unsigned int num_rel,
		   Elf_Sym *syms,
		   const char *strtab,
		   struct module *me)
{
	return -ENOEXEC;
}

static u64 calculate_toc(struct module *owner)
{
	if (!owner) {
		/* Kernel's TOC value = our current r2. */
		u64 __r2;

		asm("mr %0,2" : "=r" (__r2));
		return __r2;
	}

	return owner->arch.toc;
}

/* This is what the ELF PPC64 standard says they look like. */
struct ppc64_plt_entry
{
	void *function;
	void *r2val;
	void *unused;
};

/* Apply the given add relocations.  Return -error or 0 */
int apply_relocate_add(Elf_Rela *rela,
		       unsigned int num_rela,
		       Elf_Sym *syms,
		       const char *strtab,
		       struct module *me)
{
	unsigned long i;

	DEBUGP("Doing %u entries\n", num_rela);
	for (i = 0; i < num_rela; i++) {
		unsigned long r_type = ELF64_R_TYPE(rela[i].r_info);
		u64 *location = me->loadaddr + rela[i].r_offset;
		unsigned int symnum;
		unsigned long value = 0;
		struct module *owner;

		/* This is the symbol it is referring to.  */
		symnum = ELF64_R_SYM(rela[i].r_info);
		if (symnum) {
			value = (u64)resolve_symbol(me, symnum, syms, strtab,
						    &owner);
			if (!value)
				  return -ENOENT;
			/* This shouldn't happen. */
			if (owner == me) {
				printk("%s: Internal symbol %u referenced\n",
				       me->name, symnum);
				return -ENOEXEC;
			}
		}
		value += rela[i].r_addend;

		switch (r_type) {
		case R_PPC64_JMP_SLOT:
			/* Copy PLT entry from outside. */
			memcpy(location, (void *)value,
			       sizeof(struct ppc64_plt_entry));
			check_addr(location);
			break;

		case R_PPC64_RELATIVE:
			check_addr(location);
			*location = value + (u64)me->loadaddr;
			break;

		case R_PPC64_ADDR64:
			*location = value;
			check_addr(location);
			break;

		default:
			printk(KERN_ERR "module %s: Unknown relocation: %lu\n",
			       me->name, r_type);
			return -ENOEXEC;
		}
	}

	return 0;
}

int
module_finalize(Elf_Ehdr *hdr, struct module *me)
{
	unsigned int i;
	Elf_Shdr *sechdrs;
	char *secstrings;

	/* Find TOC value: first entry in GOT. This doesn't seem to be
           in the dynamic tags 8( */
	sechdrs = (void *)hdr + hdr->e_shoff;
	secstrings = (void *)hdr + sechdrs[hdr->e_shstrndx].sh_offset;
	for (i = 1; i < hdr->e_shnum; i++) {
		if (strcmp(secstrings+sechdrs[i].sh_name, ".got") == 0)
			me->arch.toc = *(u64 *)(me->loadaddr + 
						+ sechdrs[i].sh_addr);
	}
	return 0;
}

void module_truncate(struct module *me)
{
	/* FIXME: Implement */
}

#ifndef __KERNEL__
/* Implemented in kernsupport.S */
extern void region_cache_flush(void *addr, unsigned long size);

#define PAGE_SIZE 4096
void flush_icache_range(void *start, void *end)
{
	/* Nearest page */
	start = (void *)((unsigned long)start / PAGE_SIZE * PAGE_SIZE);
	end = (void *)ALIGN((unsigned long)end, PAGE_SIZE);

	region_cache_flush(start, end - start);
}
#endif
